
/*===============================================*\
  
  Mercredi 13 Decembre 2000  

  Arash HABIBI
  
  Medium.c
  
  Le programme principal. 
  Ce medium prend en argument un numero de port 
  (son adresse locale) et deux adresses distantes. 
  Son travail consiste a attendre derriere sa 
  socket unique. Lorsqu'il recoit un message sur 
  une des sockets, il l'envoie sur l'autre, 
  eventuellement en y inserant des erreurs, 
  eventuellement en renoncant finalement de 
  l'envoyer. 
  
\*================================================*/

#include "Medium.h"

/*------------------------------------------------*\
  Retourne le plus grand des trois nombres. 
\*------------------------------------------------*/

static int _Max(int i, int j, int k)
{
  if(j>i) 
    {
      if(k>j) return(k); 
      else    return(j);
    }
  else
    {
      if(k>i) return(k); 
      else    return(i);
    }
}

/*------------------------------------------------*\
  Cree, attache deux sockets sur deux numeros de 
  ports quelconques et affiche a l'ecran les deux 
  numeros de ports finalement trouves. 
\*------------------------------------------------*/

static void _PrepareDeuxSocketsLocaux(int *socket1_fd, int *socket2_fd)
{
  struct sockaddr_in adr_locale; 
  short port_local=0; 

  *socket1_fd = ST_GetAndBindSocket(SOCK_DGRAM, port_local, &adr_locale);
  fprintf(stderr, "\nMedium connect sur le port numro %d ", adr_locale.sin_port); 

  *socket2_fd = ST_GetAndBindSocket(SOCK_DGRAM, port_local, &adr_locale);
  fprintf(stderr, "et %d\n", adr_locale.sin_port); 
}


/*------------------------------------------------*\
  Envoie une trame contenant MEDIUM_ONLINE aux 
  deux stations. La fonction retourne M_TOUT_BAIGNE
  si tout s'est bien passe et M_ERREUR autrement. 
\*------------------------------------------------*/

static int _SendMediumOnLine(int socket_fd, struct sockaddr_in adresse)
{
  int nb_emis;
  //char *medium_on_line="MEDIUM_ON_LINE"; 
  char medium_on_line[15];
  //char *medium_on_line=""; // C'est encore plus facile

  strcpy(medium_on_line, "MEDIUM_ON_LINE");

  nb_emis = sendto(socket_fd, medium_on_line, strlen(medium_on_line)+1, 
		   0, (struct sockaddr*) &adresse, 
		   sizeof(adresse));

  if(nb_emis<0) 
    {
      if(errno==ECONNREFUSED)
	ST_CheckNetworkAddress(adresse, "_SendMediumOnLine : Apparemment personne du cote de ");
      else 
	{
	  perror("_SendMediumOnLine : sendto"); 
	  return(M_ERREUR);
	}
    }
  
  return(M_TOUT_BAIGNE); 
}


/*-------------------------------------------------------*\
  Change deux caracteres consecutifs au hasard dans chaine.
\*-------------------------------------------------------*/

static void _IntroduitErreur(char *chaine, int taille)
{
  int indice; // l'indice du caractere qu'on va changer dans la trame
  int offset; // le nombre qu'on va lui ajouter pour le changer. 

  if(taille>=0)
    {
      indice = (int)floor(drand48() * taille); 
      
      offset = (int)floor(drand48() * 256); 
      chaine[indice] = (chaine[indice]+offset)%256; 

      // pour changer aussi l'octet suivant

      if(indice+1<taille)
	{
	  offset = (int)floor(drand48() * 256); 
	  chaine[indice+1] = (chaine[indice+1]+offset)%256; 
	}
    }
}



/*------------------------------------------------*\
  On suppose qu'il y a quelque chose a lire sur 
  la socket s_rec. On lit ce quelque chose et 
  on l'envoie sur la socket s_emet vers a_dest, 
  en ayant eventuellement le soin d'y ajouter une 
  erreur ou en omettant simplement de renvoyer 
  le message. Retourne M_ERREUR en cas d'erreur
  et M_TOUT_BAIGNE autrement. 
\*------------------------------------------------*/

static int _Echange(int s_rec, int s_emet, struct sockaddr_in a_dest, float tx_prt, float tx_err)
{
  int nb_recu, nb_emis, taille_adresse; 
  char trame[M_TAILLE_TRAME]; 
  struct sockaddr_in a_expd; 

  /* ---- Lecture du message apporte par la socket ------*/

  taille_adresse = sizeof(struct sockaddr_in);
  nb_recu = recvfrom(s_rec, trame, M_TAILLE_TRAME, 
		     0, (struct sockaddr*) &a_expd, 
		     &taille_adresse);

  /* ---- Une quelconque erreur dans la lecture ---------*/

  if(nb_recu<0) 
    {
      if(errno==ECONNREFUSED) 
	{
	  //ST_CheckNetworkAddress(a_expd, "_Echange : Personne n'emet du cote de ");
	  return(M_TOUT_BAIGNE); 
	}
      else 
	{
	  perror("_Echange : recvfrom"); 
	  return(M_ERREUR);
	}
    }

  /* ---- Phase d'ecriture ----*/

  if(drand48() > tx_prt) // pour decider si la trame est perdue ou non
    {
      if(drand48() < tx_err) // pour decider si on introduit une erreur
	_IntroduitErreur(trame, nb_recu); 
      
      // --- Envoi de la trame erronnee ou non ---- */

      nb_emis = sendto(s_emet, trame, nb_recu, 
		       0, (struct sockaddr*) &a_dest, 
		       sizeof(a_dest));
      
      if(nb_emis<0) 
	{
	  if(errno==ECONNREFUSED) 
	    {
	      //ST_CheckNetworkAddress(a_dest, "_Echange : Personne n'ecoute du cote de ");
	      return(M_TOUT_BAIGNE); 
	    }
	  else 
	    {
	      perror("_Echange : sendto"); 
	      return(M_ERREUR);
	    }
	}
    }  
  return(M_TOUT_BAIGNE); 
}



/*------------------------------------------------*\
\*------------------------------------------------*/

int M_ExecMedium(struct sockaddr_in adr_1, struct sockaddr_in adr_2, float taux_perte, float taux_erreur)
{
  int socket1_fd, socket2_fd, terminal_fd, nb_fds;
  int nb_events, diagnostique=M_TOUT_BAIGNE; 
  char buffer[M_TAILLE_TRAME]; 
  fd_set lecture; 
  struct timeval timer;

  _PrepareDeuxSocketsLocaux(&socket1_fd, &socket2_fd);
  printf("Frapper \'q\' pour quitter\n\n"); 

  /*------------- Ouverture du terminal ---------------- */

  terminal_fd = open("/dev/tty", O_RDWR); 
  if(terminal_fd == -1) {perror("/dev/tty"); exit(1);}
  
  /*------- On se presente aux deux stations : --------- */

  if((_SendMediumOnLine(socket1_fd, adr_1)==M_ERREUR)||
     (_SendMediumOnLine(socket2_fd, adr_2)==M_ERREUR))
    return(M_ERREUR);

  do
    {      
      /*-------- Preparation des ensembles ------------- */
      /*------------- de descripteurs ------------------ */
  
      FD_ZERO(&lecture); 
      FD_SET( socket1_fd, &lecture); 
      FD_SET( socket2_fd, &lecture); 
      FD_SET(terminal_fd, &lecture); 

      nb_fds = 1 + _Max(terminal_fd, socket1_fd, socket2_fd);

      /*----------- Rearmement  du timer ----------------*/

      timer.tv_sec  = (int)floor(M_TIMEOUT);
      timer.tv_usec = (int)floor(1000000*(M_TIMEOUT-(int)floor(M_TIMEOUT)));

      /*------------ attente d'un evenment ------------- */

      nb_events = select (nb_fds, &lecture, NULL, NULL, &timer); 

      /*------------ Une quelconque erreur ------------- */

      if(nb_events<0)
	{perror("M_ExecMedium select"); diagnostique = M_ERREUR;}

      /*--- TIMEOUT ---*/

      else if(nb_events==0) 
	{
	  diagnostique = _SendMediumOnLine(socket1_fd, adr_1);
	  diagnostique = _SendMediumOnLine(socket2_fd, adr_2);
	}

      else // nb_events > 0
	{ 
	  /*--- Arrivee de message sur socket1 ---*/
	  if(FD_ISSET(socket1_fd, &lecture))
	    diagnostique = _Echange(socket1_fd, socket2_fd, adr_2, taux_perte, taux_erreur);

	  /*--- Arrivee de message sur socket2 ---*/
	  if(FD_ISSET(socket2_fd, &lecture))
	    diagnostique = _Echange(socket2_fd, socket1_fd, adr_1, taux_perte, taux_erreur);

	  /*--- Touche frappee au clavier -------*/
	  if(FD_ISSET(terminal_fd, &lecture))
	    {
	      if(read(terminal_fd, buffer, M_TAILLE_TRAME)==-1)
		{perror("M_ExecMedium read"); diagnostique=M_ERREUR;}
	      else 
		if(buffer[0]=='q')
		  diagnostique = M_TERMINE;
		else
		  diagnostique = M_TOUT_BAIGNE;
	    }
	}
    }
  while(diagnostique == M_TOUT_BAIGNE);
  
  return(diagnostique);
}




